#include <t_services.h>
#include <sys_config.h>
#include "kernel_id.h"
#include "485toeth.h"

#include <sys_config.h>
#include <tinet_sys_config.h>
#include <tinet_cpu_defs.h>
#include "tinet_nic_defs.h"
#include "if_ed.h"
#include <net/if.h>
#include <net/ethernet.h>
#include <net/net_buf.h>
#include <net/net_timer.h>

extern T_IF_SOFTC *sn_get_softc(void);
extern void sn_watchdog(T_IF_SOFTC *ic);
extern void sn_probe(T_IF_SOFTC *ic);
extern void sn_init(T_IF_SOFTC *ic);
extern void sn_reset(T_IF_SOFTC *ic);
extern T_NET_BUF *sn_read(T_IF_SOFTC *ic);
extern void sn_start(T_IF_SOFTC *ic, T_NET_BUF *output);

#define	EthernetCacheLength	8
#define	snCacheLength		32
#define	EthernetBuffSum		10

typedef struct {
	UB	shost[ETHER_ADDR_LEN];
	SYSTIM	stime;
} MAC_TIME;

typedef struct {
	MAC_TIME	*cachebuffer;
	UINT		cachelength;
	ID			semid_access_flag;
} MAC_CACHE;

MAC_TIME mac_time_buffer[EthernetCacheLength + snCacheLength];

MAC_CACHE EthernetCache = {
	mac_time_buffer,
	EthernetCacheLength,
	SEM_ETH_CACHE
};

MAC_CACHE snCache = {
	mac_time_buffer + EthernetCacheLength,
	snCacheLength,
	SEM_SN_CACHE
};

typedef union {
T_NET_BUF	t_net_buf;
UB			buffer[0x600];
} T_NET_BUF_NODE;

T_NET_BUF_NODE t_net_node_pool[EthernetBuffSum];

ER
tget_net_buf (T_NET_BUF **buf, UINT len, TMO tmout)
{
	UINT i;
	T_NET_BUF *p;
	ER	error = E_NOMEM;
	*buf = NULL;
	for(i = 0; i < EthernetBuffSum; i ++)
	{
		if(!t_net_node_pool[i].t_net_buf.idix)
		{
			p = &t_net_node_pool[i].t_net_buf;
			t_net_node_pool[i].t_net_buf.idix = 1;
			p->len = len;
			*buf = p;
			error = E_OK;
			break;
		}
	}
	return error;
}

BOOL isbroadcastmac(UB *d)
{
	int i = ETHER_ADDR_LEN;
	while(i--!=0)
	{
		if(*d != 0xff) return FALSE;
	}
	return TRUE;
}

inline	BOOL cmpmac(UB *d, UB *s)
{
	int i = ETHER_ADDR_LEN;
	while(i--!=0)
	{
		if(*d != *s) return FALSE;
	}
	return TRUE;
}

void	insertMACcache(UB *MAC, MAC_CACHE *p, SYSTIM stime)
{
	UINT i;
	MAC_TIME *mac_time = NULL;
	SYSTIM mintime = 0x7fffffff;
	for(i = 0; i < p->cachelength; i ++)
	{
		if(mintime > p->cachebuffer[i].stime)
		{
			mintime = p->cachebuffer[i].stime;
			mac_time = p->cachebuffer + i;
		}
		if(p->cachebuffer[i].stime && cmpmac(MAC, p->cachebuffer[i].shost))
		{
			p->cachebuffer[i].stime = stime;
			break;
		}
	}
	if(i >= p->cachelength)
	{
		wai_sem(p->semid_access_flag);
		for(i = 0; i < ETHER_ADDR_LEN; i ++)
			mac_time->shost[i] = MAC[i];
		mac_time->stime = stime;
		sig_sem(p->semid_access_flag);
	}
}

BOOL	cmpMACcache(UB *MAC, MAC_CACHE *p)
{
	UINT i;
	BOOL flag = FALSE;
	wai_sem(p->semid_access_flag);
	for(i = 0; i < p->cachelength; i ++)
	{
		if(p->cachebuffer[i].stime && cmpmac(MAC, p->cachebuffer[i].shost))
		{
			flag = TRUE;
			break;
		}
	}
	sig_sem(p->semid_access_flag);
	return flag;
}

BOOL	inEthernetCache(UB *MAC)
{
	return cmpMACcache(MAC, &EthernetCache);
}

static void
if_slowtimo (void)
{
	T_IF_SOFTC	*ic;

	ic = IF_ETHER_NIC_GET_SOFTC();

	if (ic->timer > 0) {
		ic->timer --;
		if (ic->timer == 0)
			ed_watchdog(ic);
		}

	timeout((FP)if_slowtimo, NULL, IF_TIMER_TMO);
	}

void
ifinit (void)
{
	timeout((FP)if_slowtimo, NULL, IF_TIMER_TMO);
	}


ER
sn_raw_output (T_NET_BUF *output, TMO tmout)
{
	ER error = E_OK;

	if ((error = tsnd_dtq(DTQ_SN_OUTPUT, (VP_INT)output, tmout)) != E_OK) {
		output->idix = 0;
		}
	return error;
}

ER
ether_raw_output (T_NET_BUF *output, TMO tmout)
{
	ER error = E_OK;

	if ((error = tsnd_dtq(DTQ_ETHER_OUTPUT, (VP_INT)output, tmout)) != E_OK) {
		output->idix = 0;
		}
	return error;
	}

void
my_ether_output_task(VP_INT exinf)
{
	T_IF_SOFTC	*ic;
	T_NET_BUF	*output;

	ic = IF_ETHER_NIC_GET_SOFTC();

	while (TRUE) {
		while (rcv_dtq(DTQ_ETHER_OUTPUT, (VP_INT*)(VP)&output) == E_OK) {

			syscall(wai_sem(ic->semid_txb_ready));

			IF_ETHER_NIC_START(ic, output);
			output->idix = 0;

			}
		}
	}

void
my_ether_input_task(VP_INT exinf)
{
	T_IF_SOFTC	*ic;
	T_NET_BUF	*input;
	T_ETHER_HDR	*eth;
	SYSTIM	stime;

	ifinit();

	ic = IF_ETHER_NIC_GET_SOFTC();
	IF_ETHER_NIC_PROBE(ic);
	IF_ETHER_NIC_INIT(ic);

	syscall(act_tsk(ETHER_OUTPUT_TASK));

	syscall(act_tsk(NET_TIMER_TASK));

	while (TRUE) {
		syscall(wai_sem(ic->semid_rxb_ready));
		if ((input = IF_ETHER_NIC_READ(ic)) != NULL) {
			eth = GET_ETHER_HDR(input);
			get_tim(&stime);
			insertMACcache(eth->shost, &EthernetCache, stime);
			if(isbroadcastmac(eth->dhost) || cmpMACcache(eth->dhost, &snCache))
				sn_raw_output(input, 1000);
			else input->idix = 0;
		}
	}
}


void
my_sn_output_task(VP_INT exinf)
{
	T_IF_SOFTC	*ic;
	T_NET_BUF	*output;

	ic = sn_get_softc();

	while (TRUE) {
		while (rcv_dtq(DTQ_SN_OUTPUT, (VP_INT*)(VP)&output) == E_OK) {

			syscall(wai_sem(ic->semid_txb_ready));

			sn_start(ic, output);
			output->idix = 0;

			}
		}
	}

void
my_sn_input_task(VP_INT exinf)
{
	T_IF_SOFTC	*ic;
	T_NET_BUF	*input;
	T_ETHER_HDR	*eth;
	SYSTIM	stime;

	ic = sn_get_softc();
	sn_probe(ic);
	sn_init(ic);

	syscall(act_tsk(SN_OUTPUT_TASK));

	while (TRUE) {
		syscall(wai_sem(ic->semid_rxb_ready));
		if ((input = sn_read(ic)) != NULL) {
			eth = GET_ETHER_HDR(input);
			get_tim(&stime);
			insertMACcache(eth->shost, &snCache, stime);
			ether_raw_output(input, 1000);
		}
	}
}

void main_task(VP_INT exinf)
{
	char	c;

	vmsk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG));

	syscall(serial_ctl_por(CONSOLE_PORTID,
			(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	do {
		syscall(serial_rea_dat(CONSOLE_PORTID, &c, 1));
		switch (c) {}
	} while (c != '\003' && c != 'Q');

	kernel_exit();
}
