#include <s_services.h>
#include <t_services.h>
#include "kernel_id.h"

#include <tinet_defs.h>
#include <tinet_config.h>

#include <net/if.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <net/net.h>
#include <net/net_timer.h>
#include <net/net_count.h>
#include <net/net_buf.h>

#include <sys_config.h>

/*
Data packet:
SYNC PID|/PID length Data*length CRC16
PID = 0011(DATA0)
      1011(DATA1)

Address packet:
SYNC PID|/PID length*2 TargetMAC*6 SourceMAC*6 Length*2 CRC16
PID = 1101(ADDR)

ACK packet:
SYNC PID|/PID
PID = 0010(ACK)
*/
#define	READTIMEOUT	20	/* [ms], if the time out, the new Ethernet packet begin */

#define	SYNC		0x55

#define	DATA0PACKET	0x3c
#define	DATA1PACKET	0xb4
#define	ACKPACKET	0x2d
#define	ADDRPACKET	0xd2

#define	NETSERIALSTOP	0
#define	NETSERIALSTART	1
#define	NETSERIALACK	2
#define	NETSERIALADDR	3
#define	NETSERIALDATA	4
#define	NETSERIALNAK	5

#define	SERIAL_BUFSZ	256
#define SRAMBASE		STACKTOP - 4*1024 - 32
#define	SNDPACKETLEN	128

#define INC_PTR(ptr)		{ if (++ptr == SERIAL_BUFSZ) ptr = 0; }
#define INDEX_PORT(portid)	((UINT)((portid) - 1))
#define get_nspcb(portid)	(&nspcb_table)
#define	BOUNDRATETIME 1000 * 10 / USER_BAUD_RATE

typedef struct net_serial_port_control_block {
	ID		snd_tskid;

	SYSTIM	readTime;
	SYSTIM	sndAckTime;
	SYSTIM	rcvAckTime;
	SYSTIM	sendTime;

	UH		Ethernet_rcv_packet_len;
	UH		Ethernet_snd_packet_len;

	UB		MAC[ETHER_ADDR_LEN];

	char	rcv_buffer[SERIAL_BUFSZ];
	char	Ethernet_rcv_buffer[0x600];

	UB		req_ack;
	UB		recevieFlag;
	UB		recevieChar;

	char	snd_buffer[SERIAL_BUFSZ];
	char	Ethernet_snd_buffer[0x600];
} NSPCB;

static NSPCB	nspcb_table;

typedef struct t_sn_softc {
	UW		com_number;
	} T_SN_SOFTC;

static T_SN_SOFTC sn_softc = {
	USER_PORTID
	};


T_IF_SOFTC if_softc = {
	{},
	0,
	&sn_softc,
	SEM_IF_SN_SBUF_READY,
	SEM_IF_SN_RBUF_READY,
};

#define	send_buffer()	{if(snd_write_ptr)\
	{\
		serial_wri_dat(comno, nspcb->snd_buffer, snd_write_ptr);\
		snd_write_ptr = 0;\
	}}

#define	send_char2buffer(c)		{if(snd_write_ptr >= SERIAL_BUFSZ - 1)\
		{\
			serial_wri_dat(comno, nspcb->snd_buffer, snd_write_ptr);\
			snd_write_ptr = 0;\
		}\
		if(c == SYNC || c == '\\')\
		{\
			nspcb->snd_buffer[snd_write_ptr++] = '\\';\
			c = ~c;\
		}\
		nspcb->snd_buffer[snd_write_ptr++] = c;}

const UB ackpacket[2] = {SYNC, ACKPACKET};		//return ACK packet buffer

const UH crc_ta[256]={                // CRC16=X16+X12+X5+1
0x0000, 0x1021,  0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7, 
0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef, 
0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6, 
0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de, 
0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485, 
0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d, 
0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4, 
0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc, 
0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823, 
0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b, 
0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12, 
0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a, 
0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41, 
0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49, 
0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70, 
0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78, 
0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f, 
0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067, 
0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e, 
0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256, 
0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d, 
0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405, 
0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c, 
0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634, 
0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab, 
0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3, 
0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a, 
0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92, 
0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9, 
0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1, 
0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8, 
0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0 
}; 

inline void table_crc(UB c, UH *crc)
{ 
  *crc=(*crc<<8)^crc_ta[(*crc>>8)^c];
} 

UH table_crc_len(UB *ptr, int len)
{ 
	UH crc = 0;
	while(len--!=0) 
		crc=(crc<<8)^crc_ta[(crc>>8)^*ptr++];
	return(crc);
} 

UH	seed;
UH	random()
{
	UH i;
	char *p;
	SYSTIM	stime;
	get_tim(&stime);
	p = (char*)(SRAMBASE + (seed & 0xfff));	//4k size
	for(i = 0; i < 4; i ++)
	{
		table_crc(stime, &seed);
		stime >>= 8;
	}
	for(i = 0; i < 32; i ++)
		table_crc(*p++, &seed);

	return seed;
};

void	make_MAC(UB *MAC)
{
	UH irandom;
	if(!*MAC)
	{
		MAC[0] = MAC[1] = 0x5a;
		irandom = random();
		MAC[2] = irandom;
		MAC[3] = irandom >> 8;
		irandom = random();
		MAC[4] = irandom;
		MAC[5] = irandom >> 8;
	}
}

UB	*get_MAC(ID comno, UB *MAC)
{
	UH i;
	NSPCB	*nspcb;
	nspcb = get_nspcb(comno);
	make_MAC(nspcb->MAC);
	for(i = 0; i < ETHER_ADDR_LEN; i ++) MAC[i] = nspcb->MAC[i];
	return nspcb->MAC;
}

void sndPacket_task(VP_INT exinf)
{
	T_SN_SOFTC	*sc;
	T_IF_SOFTC	*ic;

	ID	comno;

	UH i, j = 0;
	UB c, data01;
	SYSTIM	stime;

	T_SERIAL_RPOR serialstate;

	UINT	snd_write_ptr, len;
	UINT	Ethernet_snd_read_ptr;
	UH		CRC;
	NSPCB	*nspcb;

	ic = &if_softc;
	sc = ic->sc;
	comno = sc->com_number;

	nspcb = get_nspcb(comno);

	get_tim(&stime);

//wait for channel
	if(stime < nspcb->rcvAckTime) nspcb->rcvAckTime = stime;
	if(stime < nspcb->sndAckTime) nspcb->sndAckTime = stime;
	if(stime < nspcb->sendTime) nspcb->sendTime = stime;

	if((nspcb->rcvAckTime < nspcb->readTime || stime - nspcb->rcvAckTime > READTIMEOUT/2) &&
	   (nspcb->sendTime < nspcb->readTime || stime - nspcb->sendTime > READTIMEOUT/2))
	{
		nspcb->recevieFlag = 0;
		nspcb->req_ack = 0;
		nspcb->snd_tskid = TASKSND;

		if(nspcb->sndAckTime >= nspcb->readTime)
		{
			i = stime - nspcb->sndAckTime;
			if(i > READTIMEOUT) nspcb->recevieFlag = 1;
			if(i < READTIMEOUT*3/4) tslp_tsk(READTIMEOUT*3/4 - i);
		}
		else nspcb->recevieFlag = 1;

		if(nspcb->recevieFlag)
		{
			do
			{
				do{
					nspcb->recevieFlag = 0;
					tslp_tsk(READTIMEOUT*3/2);
				}while(nspcb->recevieFlag);

				c = 0;	//not send SYNC flag
				for(i = 0; i < 3; i ++)
				{
					tslp_tsk(random() & 0xf);	//wait 0-15ms
					if(nspcb->recevieFlag)
					{
						//recevie self SYNC
						if(c && nspcb->recevieChar == SYNC &&
							nspcb->readTime - nspcb->sendTime <= 5)
						{
							c = nspcb->recevieFlag = 0;
							i --;
							continue;
						}
						break;
					}
					serial_wri_dat(comno, (UB*)ackpacket, 1);
					get_tim(&nspcb->sendTime);
					c = 1;	//send SYNC flag
				}
			}while(nspcb->recevieFlag);
		}
	}

	nspcb->snd_tskid = 0;
	for(i = 0; i < ETHER_ADDR_LEN; i ++)
	{
		if(nspcb->Ethernet_snd_buffer[i] != 0xff)
		{
			nspcb->req_ack = 1;
			nspcb->snd_tskid = TASKSND;
			break;
		}
	}

	snd_write_ptr = 0;
	data01 = ADDRPACKET;
	for(Ethernet_snd_read_ptr = 0; Ethernet_snd_read_ptr < nspcb->Ethernet_snd_packet_len; Ethernet_snd_read_ptr = i)
	{
		for(j = 0; j < 3; j ++)		//try 3rd
		{
			CRC = 0;
			nspcb->snd_buffer[snd_write_ptr++] = SYNC;
			nspcb->snd_buffer[snd_write_ptr++] = data01;
			if(data01 != ADDRPACKET)
			{
				len = Ethernet_snd_read_ptr + SNDPACKETLEN;
				if(len > nspcb->Ethernet_snd_packet_len) len = nspcb->Ethernet_snd_packet_len;
				c = len - Ethernet_snd_read_ptr;
				table_crc(c, &CRC);
				send_char2buffer(c);
			}
			else
			{
				c = nspcb->Ethernet_snd_packet_len;
				table_crc(c, &CRC);
				nspcb->snd_buffer[snd_write_ptr++] = c;
				c = nspcb->Ethernet_snd_packet_len >> 8;
				table_crc(c, &CRC);
				nspcb->snd_buffer[snd_write_ptr++] = c;
				len = Ethernet_snd_read_ptr + 14;
			}
			for(i = Ethernet_snd_read_ptr; i < len; i ++)
			{
				c = nspcb->Ethernet_snd_buffer[i];
				table_crc(c, &CRC);
				send_char2buffer(c);
			}

			c = CRC >> 8;
			send_char2buffer(c);
			c = CRC;
			send_char2buffer(c);

			send_buffer();

			nspcb->recevieChar = 0;
			for(serial_ref_por(comno, &serialstate); serialstate.wricnt; serial_ref_por(comno, &serialstate))
			{
				len = serialstate.wricnt * BOUNDRATETIME;
				if(!len) break;
				tslp_tsk(len);
				if(nspcb->recevieChar == ACKPACKET) break;
			}
			if(nspcb->recevieChar == ACKPACKET) break;

			//if broadcast frame
			if(!nspcb->req_ack || !nspcb->snd_tskid)
			{
				get_tim(&nspcb->sendTime);
				nspcb->sendTime += READTIMEOUT/3;
				break;
			}

			len = READTIMEOUT/2 * BOUNDRATETIME;
			if(len < READTIMEOUT/2) len = READTIMEOUT/2;
			tslp_tsk(len);
			if(nspcb->recevieChar == ACKPACKET) break;
		}
		if(j >= 3) break;
		if(data01 == DATA0PACKET) data01 = DATA1PACKET;
		else if(data01 == DATA1PACKET) data01 = DATA0PACKET;
		else data01 = DATA0PACKET;
	}
	nspcb->snd_tskid = 0;

	if(j >= 3)
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_OUT_ERR_PACKETS], 1);

	ic->timer = 0;

	if (sig_sem(ic->semid_txb_ready) != E_OK)
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_TXB_QOVRS], 1);

	ext_tsk();
}

void rcvPacket_task(VP_INT exinf)
{
	T_SN_SOFTC	*sc;
	T_IF_SOFTC	*ic;

	ID	comno;

	UB c;					//recevie char
	UB *d, *s;

	UB isMyMAC = 0;			//2: if Ethernet packet is my MAC address; 4: broadcast MAC; 0: not my MAC

	UB data01 = 0;			//current recevie data0/1 packet flag

	UB antislashflag = 0;		//recevie a '\\'

	UB readState = NETSERIALSTOP;		//recevie state

	UINT	rcv_write_ptr = 0;
	UINT	Ethernet_rcv_write_ptr = 0;

	UH	i;

	UH		CRC;
	SYSTIM	stime;
	NSPCB	*nspcb;

	ic = &if_softc;
	sc = ic->sc;
	comno = sc->com_number;

	nspcb = get_nspcb(comno);

	while(1){
		if(readState == NETSERIALSTOP)
			isMyMAC = antislashflag = 0;

		stime = nspcb->readTime;
		serial_rea_dat(comno, &c, 1);
		get_tim(&nspcb->readTime);
		if(nspcb->snd_tskid && !nspcb->req_ack)
		{
			nspcb->recevieFlag = 1;
			nspcb->recevieChar = c;
			wup_tsk(nspcb->snd_tskid);
		}

		if(nspcb->readTime - stime > READTIMEOUT)	//read time out
			readState = NETSERIALSTOP;

		if(c == SYNC)
		{
			readState = NETSERIALSTART;
			continue;
		}
		else if(c == '\\')
		{
			if(readState == NETSERIALDATA ||
				readState == NETSERIALADDR)
			{
				antislashflag = 1;
				continue;
			}
			readState = NETSERIALSTOP;
		}

		if(antislashflag)
		{
			c = ~c;
			if(c != SYNC && c != '\\')
				readState = NETSERIALSTOP;
		}
		antislashflag = 0;

		switch(readState)
		{
			case NETSERIALSTART:
				CRC = 0;
				rcv_write_ptr = 0;
				switch(c)	//PID
				{
					case DATA0PACKET:
						if(isMyMAC)
						{
							data01 = 0;
							readState = NETSERIALDATA;
						}
						break;
					case DATA1PACKET:
						if(isMyMAC)
						{
							data01 = 1;
							readState = NETSERIALDATA;
						}
						break;

					case ADDRPACKET:
						readState = NETSERIALADDR;
						break;

					case ACKPACKET:
						if(nspcb->snd_tskid && nspcb->req_ack)
						{
							get_tim(&nspcb->rcvAckTime);
							nspcb->recevieChar = c;
							wup_tsk(nspcb->snd_tskid);
						}
						break;

					default:
						readState = NETSERIALSTOP;
						break;
				};
				break;

			case NETSERIALADDR:
				nspcb->rcv_buffer[rcv_write_ptr] = c;
				INC_PTR(rcv_write_ptr);
				table_crc(c, &CRC);
				if(rcv_write_ptr >= 18)
				{
					if(!CRC)
					{
						d = nspcb->Ethernet_rcv_buffer;
						s = nspcb->rcv_buffer;

						nspcb->Ethernet_rcv_packet_len = s[1];
						nspcb->Ethernet_rcv_packet_len <<= 8;
						nspcb->Ethernet_rcv_packet_len += *s;
						s += 2;

						for(i = 0; i < 14; i ++) *d++ = *s++;
						d = nspcb->Ethernet_rcv_buffer;
						s = nspcb->MAC;
						isMyMAC = 2;
						for(i = 0; i < ETHER_ADDR_LEN; i ++)
						{
							if(*d++ != *s++) {isMyMAC = 0;break;}
						}
						if(!isMyMAC)
						{
							isMyMAC = 4;
							d = nspcb->Ethernet_rcv_buffer;
							for(i = 0; i < ETHER_ADDR_LEN; i ++)
							{
								if(*d++ != 0xff) {isMyMAC = 0; break;}
							}
						}

						if(isMyMAC)
						{
							Ethernet_rcv_write_ptr = 14;
							if(nspcb->Ethernet_rcv_packet_len > 1514 ||
								nspcb->Ethernet_rcv_packet_len < 14) isMyMAC = 0;

							if(isMyMAC & 2)	//if not broadcast, no ack. 2: my mac; 4: broadcast
							{
								serial_wri_dat(comno, (UB*)ackpacket, 2);
								get_tim(&nspcb->sndAckTime);
								nspcb->sndAckTime += READTIMEOUT/3;
							}
						}
					}
					if(!isMyMAC)	//error
						readState = NETSERIALSTOP;
				}
				break;

			case NETSERIALDATA:
				nspcb->rcv_buffer[rcv_write_ptr] = c;
				INC_PTR(rcv_write_ptr);
				table_crc(c, &CRC);
				if(!CRC && rcv_write_ptr == *nspcb->rcv_buffer + 3)
				{
					rcv_write_ptr -= 3;
					if(rcv_write_ptr <= nspcb->Ethernet_rcv_packet_len - Ethernet_rcv_write_ptr)
					{
						if((isMyMAC & 1) == data01)
						{
							d = nspcb->Ethernet_rcv_buffer + Ethernet_rcv_write_ptr;
							s = nspcb->rcv_buffer + 1;
							i = rcv_write_ptr;
							while(i--!=0) *d++ = *s++;
							Ethernet_rcv_write_ptr += rcv_write_ptr;
							isMyMAC ^= 1;
						}

						if(isMyMAC & 2)	//if not broadcast, no ack. 2: my mac; 4: broadcast
						{
							serial_wri_dat(comno, (UB*)ackpacket, 2);
							get_tim(&nspcb->sndAckTime);
							nspcb->sndAckTime += READTIMEOUT/3;
						}
						if(Ethernet_rcv_write_ptr == nspcb->Ethernet_rcv_packet_len)
						{
							if (sig_sem(ic->semid_rxb_ready) != E_OK)
								NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_RXB_QOVRS], 1);
						}
					}
				}
				break;
			
			default:
				readState = NETSERIALSTOP;
				break;
		}
	}
}

T_IF_SOFTC *
sn_get_softc (void)
{
	return &if_softc;
	}

void
sn_probe (T_IF_SOFTC *ic)
{
	T_SN_SOFTC	*sc;
	ID	comno;
	sc = ic->sc;
	comno = sc->com_number;

	syscall(serial_opn_por(comno));
	syscall(serial_ctl_por(comno, 0));
	get_MAC(comno, ic->ifaddr.lladdr);
	act_tsk(TASKRCV);
}

void
sn_init (T_IF_SOFTC *ic)
{
	ic->timer = 0;
}

void
sn_reset (T_IF_SOFTC *ic)
{
	NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_RESETS], 1);
	sn_init(ic);
}

void
sn_watchdog (T_IF_SOFTC *ic)
{
	NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_OUT_ERR_PACKETS], 1);
	NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_TIMEOUTS], 1);
	sn_reset(ic);
	}
/*
inline	UB	*char2hex(UB c, UB *p)
{
	UB a;
	*p ++ = '0';
	*p ++ = 'x';
	a = c >> 4;
	if(a <= 9) *p ++ = '0' + a;
	else *p ++ = 'a' - 10 + a;
	a = c & 0xf;
	if(a <= 9) *p ++ = '0' + a;
	else *p ++ = 'a' - 10 + a;
	*p ++ = ',';
	return p;
}
*/
void
sn_start (T_IF_SOFTC *ic, T_NET_BUF *output)
{
	UH i;
	UB *d, *s;
	NSPCB	*nspcb;
	nspcb = get_nspcb(USER_PORTID);

	NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_OUT_PACKETS], 1);
	NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_OUT_OCTETS],  output->len);

//	syslog(LOG_NOTICE, "TD length = %d", output->len);
//	syslog(LOG_NOTICE, "Target %x:%x:%x", output->buf[0], output->buf[1], output->buf[2]);
//	syslog(LOG_NOTICE, "Target %x:%x:%x", output->buf[3], output->buf[4], output->buf[5]);
//	syslog(LOG_NOTICE, "Source %x:%x:%x", output->buf[6], output->buf[7], output->buf[8]);
//	syslog(LOG_NOTICE, "Source %x:%x:%x", output->buf[9], output->buf[10], output->buf[11]);
//	syslog(LOG_NOTICE, "Type %x:%x", output->buf[12], output->buf[13]);

	d = nspcb->Ethernet_snd_buffer;
	s = output->buf;
	nspcb->Ethernet_snd_packet_len = i = output->len;
	while(i--!=0) *d++ = *s++;

/*
	if(nspcb->Ethernet_snd_packet_len == 106 && output->buf[12] == 8 && output->buf[13] == 0)
	{
		i = 106;
		s = output->buf;
		d = (UB *)0x480000;
		while(i--!=0)
			d = char2hex(*s++, d);
		*d = 0;
		syslog(LOG_NOTICE, (UB *)0x480000);
	}*/
	act_tsk(TASKSND);
}

T_NET_BUF *
sn_read (T_IF_SOFTC *ic)
{
	UB *d, *s;
	T_NET_BUF	*input = NULL;
	UH		align;
	int		len;
	NSPCB	*nspcb;
	nspcb = get_nspcb(USER_PORTID);

	len = nspcb->Ethernet_rcv_packet_len;

	align = ((((len - sizeof(T_IF_HDR)) + 3) >> 2) << 2) + sizeof(T_IF_HDR);

	if (tget_net_buf(&input, align, TMO_IF_SN_GET_NET_BUF) == E_OK && input != NULL) {
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_IN_PACKETS], 1);
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_IN_OCTETS],  len);
		d = input->buf;
		s = nspcb->Ethernet_rcv_buffer;
		while(len--!=0) *d++ = *s++;
//	syslog(LOG_NOTICE, "RD length = %d", nspcb->Ethernet_rcv_packet_len);
//	syslog(LOG_NOTICE, "Target %x:%x:%x", input->buf[0], input->buf[1], input->buf[2]);
//	syslog(LOG_NOTICE, "Target %x:%x:%x", input->buf[3], input->buf[4], input->buf[5]);
//	syslog(LOG_NOTICE, "Source %x:%x:%x", input->buf[6], input->buf[7], input->buf[8]);
//	syslog(LOG_NOTICE, "Source %x:%x:%x", input->buf[9], input->buf[10], input->buf[11]);
//	syslog(LOG_NOTICE, "Type %x:%x", input->buf[12], input->buf[13]);
//	syslog(LOG_NOTICE, "Align length = %d", input->len);
		}
	else {
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_IN_ERR_PACKETS], 1);
		NET_COUNT_ETHER_NIC(net_count_ether_nic[NC_ETHER_NIC_NO_BUFS], 1);
		}
	return input;
}

